@@ -1,12 +1,11 @@
 use crate::internal::{
-    self, consts, Allocator, Chain, DirEntry, Entries, EntriesOrder, ObjType,
-    Sector, SectorInit, Version,
+    self, consts, Allocator, Chain, Color, DirEntry, ObjType, Sector,
+    SectorInit, Timestamp, Validation, Version,
 };
 use byteorder::{LittleEndian, WriteBytesExt};
+use fnv::FnvHashSet;
 use std::cmp::Ordering;
-use std::collections::HashSet;
 use std::io::{self, Seek, SeekFrom, Write};
-use std::path::Path;
 
 //===========================================================================//
 
@@ -32,9 +31,10 @@ impl<F> Directory<F> {
         allocator: Allocator<F>,
         dir_entries: Vec<DirEntry>,
         dir_start_sector: u32,
+        validation: Validation,
     ) -> io::Result<Directory<F>> {
         let directory = Directory { allocator, dir_entries, dir_start_sector };
-        directory.validate()?;
+        directory.validate(validation)?;
         Ok(directory)
     }
 
@@ -42,6 +42,10 @@ impl<F> Directory<F> {
         self.allocator.version()
     }
 
+    pub fn inner(&self) -> &F {
+        self.allocator.inner()
+    }
+
     pub fn sector_len(&self) -> usize {
         self.allocator.sector_len()
     }
@@ -59,7 +63,7 @@ impl<F> Directory<F> {
                     return None;
                 }
                 let dir_entry = self.dir_entry(stream_id);
-                match internal::path::compare_names(&name, &dir_entry.name) {
+                match internal::path::compare_names(name, &dir_entry.name) {
                     Ordering::Equal => break,
                     Ordering::Less => stream_id = dir_entry.left_sibling,
                     Ordering::Greater => stream_id = dir_entry.right_sibling,
@@ -69,85 +73,11 @@ impl<F> Directory<F> {
         Some(stream_id)
     }
 
-    /// Returns an iterator over the entries within the root storage object.
-    pub fn root_storage_entries(&self) -> Entries {
-        let start = self.root_dir_entry().child;
-        Entries::new(
-            EntriesOrder::Nonrecursive,
-            &self.dir_entries,
-            internal::path::path_from_name_chain(&[]),
-            start,
-        )
-    }
-
-    /// Returns an iterator over the entries within a storage object.
-    pub fn storage_entries(&self, path: &Path) -> io::Result<Entries> {
-        let names = internal::path::name_chain_from_path(path)?;
-        let path = internal::path::path_from_name_chain(&names);
-        let stream_id = match self.stream_id_for_name_chain(&names) {
-            Some(stream_id) => stream_id,
-            None => not_found!("No such storage: {:?}", path),
-        };
-        let start = {
-            let dir_entry = self.dir_entry(stream_id);
-            if dir_entry.obj_type == ObjType::Stream {
-                invalid_input!("Not a storage: {:?}", path);
-            }
-            debug_assert!(
-                dir_entry.obj_type == ObjType::Storage
-                    || dir_entry.obj_type == ObjType::Root
-            );
-            dir_entry.child
-        };
-        Ok(Entries::new(
-            EntriesOrder::Nonrecursive,
-            &self.dir_entries,
-            path,
-            start,
-        ))
-    }
-
-    /// Returns an iterator over all entries within the compound file, starting
-    /// from and including the root entry.  The iterator walks the storage tree
-    /// in a preorder traversal.
-    pub fn walk(&self) -> Entries {
-        Entries::new(
-            EntriesOrder::Preorder,
-            &self.dir_entries,
-            internal::path::path_from_name_chain(&[]),
-            consts::ROOT_STREAM_ID,
-        )
-    }
-
-    /// Returns an iterator over all entries under a storage subtree, including
-    /// the given path itself.  The iterator walks the storage tree in a
-    /// preorder traversal.
-    pub fn walk_storage(&self, path: &Path) -> io::Result<Entries> {
-        let mut names = internal::path::name_chain_from_path(path)?;
-        let stream_id = match self.stream_id_for_name_chain(&names) {
-            Some(stream_id) => stream_id,
-            None => {
-                not_found!(
-                    "No such object: {:?}",
-                    internal::path::path_from_name_chain(&names)
-                );
-            }
-        };
-        names.pop();
-        let parent_path = internal::path::path_from_name_chain(&names);
-        Ok(Entries::new(
-            EntriesOrder::Preorder,
-            &self.dir_entries,
-            parent_path,
-            stream_id,
-        ))
-    }
-
     pub fn open_chain(
         &mut self,
         start_sector_id: u32,
         init: SectorInit,
-    ) -> Chain<F> {
+    ) -> io::Result<Chain<F>> {
         self.allocator.open_chain(start_sector_id, init)
     }
 
@@ -163,15 +93,11 @@ impl<F> Directory<F> {
         &mut self.dir_entries[stream_id as usize]
     }
 
-    fn validate(&self) -> io::Result<()> {
-        let root_entry = self.root_dir_entry();
-        if root_entry.name != consts::ROOT_DIR_NAME {
-            malformed!(
-                "root entry name is {:?}, but should be {:?}",
-                root_entry.name,
-                consts::ROOT_DIR_NAME
-            );
+    fn validate(&self, validation: Validation) -> io::Result<()> {
+        if self.dir_entries.is_empty() {
+            malformed!("root entry is missing");
         }
+        let root_entry = self.root_dir_entry();
         if root_entry.stream_len % consts::MINI_SECTOR_LEN as u64 != 0 {
             malformed!(
                 "root stream len is {}, but should be multiple of {}",
@@ -179,9 +105,9 @@ impl<F> Directory<F> {
                 consts::MINI_SECTOR_LEN
             );
         }
-        let mut visited = HashSet::new();
-        let mut stack = vec![consts::ROOT_STREAM_ID];
-        while let Some(stream_id) = stack.pop() {
+        let mut visited = FnvHashSet::default();
+        let mut stack = vec![(consts::ROOT_STREAM_ID, false)];
+        while let Some((stream_id, parent_is_red)) = stack.pop() {
             if visited.contains(&stream_id) {
                 malformed!("loop in tree");
             }
@@ -202,6 +128,16 @@ impl<F> Directory<F> {
                     dir_entry.obj_type
                 );
             }
+            let node_is_red = dir_entry.color == Color::Red;
+            // The MS-CFB spec section 2.6.4 says that two consecutive nodes in
+            // the red-black tree for siblings within a storage object MUST NOT
+            // both be red, but apparently some implementations don't obey this
+            // (see https://github.com/mdsteele/rust-cfb/issues/10).  We still
+            // want to be able to read these files, so we only consider this an
+            // error under Strict validation.
+            if parent_is_red && node_is_red && validation.is_strict() {
+                malformed!("RB tree has adjacent red nodes");
+            }
             let left_sibling = dir_entry.left_sibling;
             if left_sibling != consts::NO_STREAM {
                 if left_sibling as usize >= self.dir_entries.len() {
@@ -222,7 +158,7 @@ impl<F> Directory<F> {
                         entry.name
                     );
                 }
-                stack.push(left_sibling);
+                stack.push((left_sibling, node_is_red));
             }
             let right_sibling = dir_entry.right_sibling;
             if right_sibling != consts::NO_STREAM {
@@ -242,7 +178,7 @@ impl<F> Directory<F> {
                         entry.name
                     );
                 }
-                stack.push(right_sibling);
+                stack.push((right_sibling, node_is_red));
             }
             let child = dir_entry.child;
             if child != consts::NO_STREAM {
@@ -253,7 +189,7 @@ impl<F> Directory<F> {
                         self.dir_entries.len()
                     );
                 }
-                stack.push(child);
+                stack.push((child, false));
             }
         }
         Ok(())
@@ -283,7 +219,7 @@ impl<F: Seek> Directory<F> {
         let mut directory_sector = self.dir_start_sector;
         for _ in 0..(stream_id / dir_entries_per_sector) {
             debug_assert_ne!(directory_sector, consts::END_OF_CHAIN);
-            directory_sector = self.allocator.next(directory_sector);
+            directory_sector = self.allocator.next(directory_sector)?;
         }
         self.allocator.seek_within_subsector(
             directory_sector,
@@ -330,8 +266,12 @@ impl<F: Write + Seek> Directory<F> {
         );
         // Create a new directory entry.
         let stream_id = self.allocate_dir_entry()?;
-        let now = internal::time::current_timestamp();
-        *self.dir_entry_mut(stream_id) = DirEntry::new(name, obj_type, now);
+        // 2.6.1 streams must have creation and modified time of 0
+        let mut ts = Timestamp::zero();
+        if obj_type == ObjType::Storage {
+            ts = Timestamp::now();
+        }
+        *self.dir_entry_mut(stream_id) = DirEntry::new(name, obj_type, ts);
 
         // Insert the new entry into the tree.
         let mut sibling_id = self.dir_entry(parent_id).child;
@@ -517,8 +457,9 @@ impl<F: Write + Seek> Directory<F> {
     }
 
     fn write_dir_entry(&mut self, stream_id: u32) -> io::Result<()> {
-        let mut chain =
-            self.allocator.open_chain(self.dir_start_sector, SectorInit::Dir);
+        let mut chain = self
+            .allocator
+            .open_chain(self.dir_start_sector, SectorInit::Dir)?;
         let offset = (consts::DIR_ENTRY_LEN as u64) * (stream_id as u64);
         chain.seek(SeekFrom::Start(offset))?;
         self.dir_entries[stream_id as usize].write_to(&mut chain)
@@ -536,11 +477,15 @@ impl<F: Write + Seek> Directory<F> {
 mod tests {
     use super::Directory;
     use crate::internal::{
-        consts, Allocator, Color, DirEntry, ObjType, Sectors, Version,
+        consts, Allocator, Color, DirEntry, ObjType, Sectors, Timestamp,
+        Validation, Version,
     };
     use std::io::Cursor;
 
-    fn make_directory(entries: Vec<DirEntry>) -> Directory<Cursor<Vec<u8>>> {
+    fn make_directory(
+        entries: Vec<DirEntry>,
+        validation: Validation,
+    ) -> Directory<Cursor<Vec<u8>>> {
         let version = Version::V3;
         let num_sectors = 3;
         let data_len = (1 + num_sectors) * version.sector_len();
@@ -548,19 +493,15 @@ mod tests {
         let sectors = Sectors::new(version, data_len as u64, cursor);
         let mut fat = vec![consts::END_OF_CHAIN; num_sectors];
         fat[0] = consts::FAT_SECTOR;
-        let allocator = Allocator::new(sectors, vec![], vec![0], fat).unwrap();
-        Directory::new(allocator, entries, 1).unwrap()
+        let allocator =
+            Allocator::new(sectors, vec![], vec![0], fat, validation).unwrap();
+        Directory::new(allocator, entries, 1, validation).unwrap()
     }
 
     #[test]
-    #[should_panic(
-        expected = "Malformed directory (root entry name is \\\"foo\\\", but \
-                    should be \\\"Root Entry\\\")"
-    )]
-    fn incorrect_root_entry_name() {
-        let mut root_entry = DirEntry::empty_root_entry();
-        root_entry.name = "foo".to_string();
-        make_directory(vec![root_entry]);
+    #[should_panic(expected = "Malformed directory (root entry is missing)")]
+    fn no_root_entry() {
+        make_directory(vec![], Validation::Permissive);
     }
 
     #[test]
@@ -572,7 +513,7 @@ mod tests {
         let mut root_entry = DirEntry::empty_root_entry();
         root_entry.start_sector = 2;
         root_entry.stream_len = 147;
-        make_directory(vec![root_entry]);
+        make_directory(vec![root_entry], Validation::Permissive);
     }
 
     #[test]
@@ -580,9 +521,10 @@ mod tests {
     fn storage_is_child_of_itself() {
         let mut root_entry = DirEntry::empty_root_entry();
         root_entry.child = 1;
-        let mut storage = DirEntry::new("foo", ObjType::Storage, 0);
+        let mut storage =
+            DirEntry::new("foo", ObjType::Storage, Timestamp::zero());
         storage.child = 1;
-        make_directory(vec![root_entry, storage]);
+        make_directory(vec![root_entry, storage], Validation::Permissive);
     }
 
     #[test]
@@ -592,7 +534,7 @@ mod tests {
     fn root_has_wrong_type() {
         let mut root_entry = DirEntry::empty_root_entry();
         root_entry.obj_type = ObjType::Storage;
-        make_directory(vec![root_entry]);
+        make_directory(vec![root_entry], Validation::Permissive);
     }
 
     #[test]
@@ -602,8 +544,8 @@ mod tests {
     fn nonroot_has_wrong_type() {
         let mut root_entry = DirEntry::empty_root_entry();
         root_entry.child = 1;
-        let storage = DirEntry::new("foo", ObjType::Root, 0);
-        make_directory(vec![root_entry, storage]);
+        let storage = DirEntry::new("foo", ObjType::Root, Timestamp::zero());
+        make_directory(vec![root_entry, storage], Validation::Permissive);
     }
 
     #[test]
@@ -615,24 +557,39 @@ mod tests {
         // we shouldn't complain if the root is red.
         let mut root_entry = DirEntry::empty_root_entry();
         root_entry.color = Color::Red;
-        make_directory(vec![root_entry]);
+        make_directory(vec![root_entry], Validation::Permissive);
+    }
+
+    fn make_entries_with_adjacent_red_nodes() -> Vec<DirEntry> {
+        let mut root_entry = DirEntry::empty_root_entry();
+        root_entry.child = 1;
+        let mut storage1 =
+            DirEntry::new("foo", ObjType::Storage, Timestamp::zero());
+        storage1.color = Color::Red;
+        storage1.left_sibling = 2;
+        let mut storage2 =
+            DirEntry::new("bar", ObjType::Storage, Timestamp::zero());
+        storage2.color = Color::Red;
+        vec![root_entry, storage1, storage2]
     }
 
     #[test]
-    fn tolerate_two_red_nodes_in_a_row() {
-        // The MS-CFB spec section 2.6.4 says that two consecutive nodes in the
-        // tree MUST NOT both be red, but apparently some implementations don't
-        // obey this (see https://github.com/mdsteele/rust-cfb/issues/10).  We
-        // still want to be able to read these files, so we shouldn't complain
-        // if there are two red nodes in a row.
-        let mut root_entry = DirEntry::empty_root_entry();
-        root_entry.child = 1;
-        let mut storage1 = DirEntry::new("foo", ObjType::Storage, 0);
-        storage1.color = Color::Red;
-        storage1.left_sibling = 2;
-        let mut storage2 = DirEntry::new("bar", ObjType::Storage, 0);
-        storage2.color = Color::Red;
-        make_directory(vec![root_entry, storage1, storage2]);
+    #[should_panic(
+        expected = "Malformed directory (RB tree has adjacent red nodes)"
+    )]
+    fn adjacent_red_nodes_strict() {
+        make_directory(
+            make_entries_with_adjacent_red_nodes(),
+            Validation::Strict,
+        );
+    }
+
+    #[test]
+    fn adjacent_red_nodes_permissive() {
+        make_directory(
+            make_entries_with_adjacent_red_nodes(),
+            Validation::Permissive,
+        );
     }
 }
 
