@@ -1,5 +1,6 @@
-use cfb::{CompoundFile, Version};
-use std::io::{Cursor, Read, Seek, SeekFrom, Write};
+use cfb::{CompoundFile, Entry, Version};
+use std::io::{self, Cursor, Read, Seek, SeekFrom, Write};
+use std::path::Path;
 use uuid::Uuid;
 
 //===========================================================================//
@@ -12,15 +13,8 @@ fn read_storage_to_vec<F>(comp: &CompoundFile<F>, path: &str) -> Vec<String> {
     comp.read_storage(path).unwrap().map(|e| e.name().to_string()).collect()
 }
 
-fn walk_to_vec<F>(comp: &CompoundFile<F>) -> Vec<String> {
-    comp.walk().map(|e| e.path().to_string_lossy().to_string()).collect()
-}
-
-fn walk_storage_to_vec<F>(comp: &CompoundFile<F>, path: &str) -> Vec<String> {
-    comp.walk_storage(path)
-        .unwrap()
-        .map(|e| e.path().to_string_lossy().to_string())
-        .collect()
+fn walk_to_vec(entries: &[Entry]) -> Vec<&Path> {
+    entries.iter().map(|e| e.path()).collect()
 }
 
 //===========================================================================//
@@ -45,7 +39,7 @@ fn create_empty_compound_file() {
 
     let cursor = comp.into_inner();
     assert_eq!(cursor.get_ref().len(), 3 * version.sector_len());
-    let comp = CompoundFile::open(cursor).expect("open");
+    let comp = CompoundFile::open_strict(cursor).expect("open");
     assert_eq!(comp.version(), version);
     assert_eq!(comp.entry("/").unwrap().name(), "Root Entry");
 }
@@ -120,7 +114,7 @@ fn create_directory_tree() {
     comp.create_storage("/foo/bar").unwrap();
 
     let cursor = comp.into_inner();
-    let comp = CompoundFile::open(cursor).expect("open");
+    let comp = CompoundFile::open_strict(cursor).expect("open");
     assert_eq!(read_root_storage_to_vec(&comp), vec!["baz", "foo"]);
     assert_eq!(read_storage_to_vec(&comp, "/"), vec!["baz", "foo"]);
     assert_eq!(read_storage_to_vec(&comp, "/foo"), vec!["bar"]);
@@ -136,16 +130,35 @@ fn walk_directory_tree() {
     comp.create_stream("/baz").unwrap();
     comp.create_storage("/quux").unwrap();
     comp.create_stream("/foo/bar").unwrap();
+    let entries: Vec<Entry> = comp.walk().collect();
     assert_eq!(
-        walk_to_vec(&comp),
-        vec!["/", "/baz", "/foo", "/foo/bar", "/quux"]
+        walk_to_vec(&entries),
+        vec![
+            Path::new("/"),
+            Path::new("/baz"),
+            Path::new("/foo"),
+            Path::new("/foo/bar"),
+            Path::new("/quux")
+        ]
     );
+    let entries: Vec<Entry> = comp.walk_storage("/").unwrap().collect();
     assert_eq!(
-        walk_storage_to_vec(&comp, "/"),
-        vec!["/", "/baz", "/foo", "/foo/bar", "/quux"]
+        walk_to_vec(&entries),
+        vec![
+            Path::new("/"),
+            Path::new("/baz"),
+            Path::new("/foo"),
+            Path::new("/foo/bar"),
+            Path::new("/quux")
+        ]
     );
-    assert_eq!(walk_storage_to_vec(&comp, "/foo"), vec!["/foo", "/foo/bar"]);
-    assert_eq!(walk_storage_to_vec(&comp, "/baz"), vec!["/baz"]);
+    let entries: Vec<Entry> = comp.walk_storage("/foo").unwrap().collect();
+    assert_eq!(
+        walk_to_vec(&entries),
+        vec![Path::new("/foo"), Path::new("/foo/bar")]
+    );
+    let entries: Vec<Entry> = comp.walk_storage("/baz").unwrap().collect();
+    assert_eq!(walk_to_vec(&entries), vec![Path::new("/baz")]);
 }
 
 #[test]
@@ -243,7 +256,7 @@ fn storage_clsids() {
     assert_eq!(comp.entry("/foo").unwrap().clsid(), &uuid2);
 
     let cursor = comp.into_inner();
-    let comp = CompoundFile::open(cursor).expect("open");
+    let comp = CompoundFile::open_strict(cursor).expect("open");
     assert_eq!(comp.root_entry().clsid(), &uuid1);
     assert_eq!(comp.entry("/foo").unwrap().clsid(), &uuid2);
 }
@@ -282,7 +295,7 @@ fn state_bits() {
     assert_eq!(comp.entry("bar").unwrap().state_bits(), 0x0ABCDEF0);
 
     let cursor = comp.into_inner();
-    let comp = CompoundFile::open(cursor).expect("open");
+    let comp = CompoundFile::open_strict(cursor).expect("open");
     assert_eq!(comp.root_entry().state_bits(), 0);
     assert_eq!(comp.entry("foo").unwrap().state_bits(), 0x12345678);
     assert_eq!(comp.entry("bar").unwrap().state_bits(), 0x0ABCDEF0);
@@ -344,8 +357,7 @@ fn create_storages_all() {
 }
 
 #[test]
-#[should_panic(expected = "Cannot create storage at \\\"/foo/bar\\\" \
-                           because a stream already exists there")]
+#[should_panic(expected = "Cannot create storage")]
 fn create_storage_all_with_stream_in_the_way() {
     let cursor = Cursor::new(Vec::new());
     let mut comp = CompoundFile::create(cursor).expect("create");
@@ -371,7 +383,7 @@ fn remove_storages() {
     comp.remove_storage("/baz/blarg").unwrap();
 
     let cursor = comp.into_inner();
-    let comp = CompoundFile::open(cursor).expect("open");
+    let comp = CompoundFile::open_strict(cursor).expect("open");
     assert_eq!(read_storage_to_vec(&comp, "/"), vec!["baz", "quux"]);
     assert!(read_storage_to_vec(&comp, "/baz").is_empty());
 }
@@ -425,7 +437,7 @@ fn remove_storage_all_on_storage() {
     comp.remove_storage_all("foo").unwrap();
 
     let cursor = comp.into_inner();
-    let comp = CompoundFile::open(cursor).expect("open");
+    let comp = CompoundFile::open_strict(cursor).expect("open");
     assert_eq!(read_storage_to_vec(&comp, "/"), vec!["stuff"]);
     assert_eq!(read_storage_to_vec(&comp, "/stuff"), vec!["foo"]);
 }
@@ -441,7 +453,7 @@ fn remove_storage_all_on_root() {
     comp.remove_storage_all("/").unwrap();
 
     let cursor = comp.into_inner();
-    let comp = CompoundFile::open(cursor).expect("open");
+    let comp = CompoundFile::open_strict(cursor).expect("open");
     assert!(read_storage_to_vec(&comp, "/").is_empty());
 }
 
@@ -456,7 +468,7 @@ fn create_streams() {
     comp.create_stream("/baz").unwrap().write_all(b"baz!").unwrap();
 
     let cursor = comp.into_inner();
-    let mut comp = CompoundFile::open(cursor).expect("open");
+    let mut comp = CompoundFile::open_strict(cursor).expect("open");
     {
         let mut stream = comp.open_stream("/foo").unwrap();
         let mut data = String::new();
@@ -481,7 +493,7 @@ fn create_small_stream() {
     comp.create_stream("foobar").unwrap().write_all(&data).unwrap();
 
     let cursor = comp.into_inner();
-    let mut comp = CompoundFile::open(cursor).expect("open");
+    let mut comp = CompoundFile::open_strict(cursor).expect("open");
     let mut stream = comp.open_stream("foobar").unwrap();
     let mut actual_data = Vec::new();
     stream.read_to_end(&mut actual_data).unwrap();
@@ -498,7 +510,7 @@ fn create_large_stream() {
     comp.create_stream("foobar").unwrap().write_all(&data).unwrap();
 
     let cursor = comp.into_inner();
-    let mut comp = CompoundFile::open(cursor).expect("open");
+    let mut comp = CompoundFile::open_strict(cursor).expect("open");
     let mut stream = comp.open_stream("foobar").unwrap();
     let mut actual_data = Vec::new();
     stream.read_to_end(&mut actual_data).unwrap();
@@ -519,7 +531,7 @@ fn create_very_large_stream() {
     }
 
     let cursor = comp.into_inner();
-    let mut comp = CompoundFile::open(cursor).expect("open");
+    let mut comp = CompoundFile::open_strict(cursor).expect("open");
     let mut stream = comp.open_stream("foobar").unwrap();
     assert_eq!(stream.len(), 1_000_000);
     assert_eq!(stream.seek(SeekFrom::End(0)).unwrap(), 1_000_000);
@@ -586,7 +598,7 @@ fn remove_streams() {
     comp.remove_stream("/baz/blarg").unwrap();
 
     let cursor = comp.into_inner();
-    let comp = CompoundFile::open(cursor).expect("open");
+    let comp = CompoundFile::open_strict(cursor).expect("open");
     assert_eq!(read_storage_to_vec(&comp, "/"), vec!["baz", "quux"]);
     assert!(read_storage_to_vec(&comp, "/baz").is_empty());
 }
@@ -631,7 +643,7 @@ fn truncate_stream() {
         assert_eq!(stream.seek(SeekFrom::Start(6000)).unwrap(), 6000);
         stream.set_len(7000).unwrap();
         assert_eq!(stream.len(), 7000);
-        assert_eq!(stream.seek(SeekFrom::Current(0)).unwrap(), 6000);
+        assert_eq!(stream.stream_position().unwrap(), 6000);
         stream.set_len(5000).unwrap();
         assert_eq!(stream.len(), 5000);
         stream.write_all(&vec![b'x'; 1000]).unwrap();
@@ -639,7 +651,7 @@ fn truncate_stream() {
     }
 
     let cursor = comp.into_inner();
-    let mut comp = CompoundFile::open(cursor).expect("open");
+    let mut comp = CompoundFile::open_strict(cursor).expect("open");
     let mut stream = comp.open_stream("/foobar").unwrap();
     assert_eq!(stream.len(), 6000);
     let mut actual_data = Vec::new();
@@ -660,17 +672,17 @@ fn extend_stream() {
         assert_eq!(stream.seek(SeekFrom::Start(1000)).unwrap(), 1000);
         stream.write_all(&vec![b'y'; 500]).unwrap();
         assert_eq!(stream.len(), 2000);
-        assert_eq!(stream.seek(SeekFrom::Current(0)).unwrap(), 1500);
+        assert_eq!(stream.stream_position().unwrap(), 1500);
         stream.set_len(5000).unwrap();
         assert_eq!(stream.len(), 5000);
-        assert_eq!(stream.seek(SeekFrom::Current(0)).unwrap(), 1500);
+        assert_eq!(stream.stream_position().unwrap(), 1500);
         stream.write_all(&vec![b'z'; 500]).unwrap();
         assert_eq!(stream.len(), 5000);
-        assert_eq!(stream.seek(SeekFrom::Current(0)).unwrap(), 2000);
+        assert_eq!(stream.stream_position().unwrap(), 2000);
     }
 
     let cursor = comp.into_inner();
-    let mut comp = CompoundFile::open(cursor).expect("open");
+    let mut comp = CompoundFile::open_strict(cursor).expect("open");
     let mut stream = comp.open_stream("/foobar").unwrap();
     assert_eq!(stream.len(), 5000);
     let mut actual_data = Vec::new();
@@ -708,3 +720,52 @@ fn stream_seek() {
 }
 
 //===========================================================================//
+// Tests for opening multiple streams at once:
+
+#[test]
+fn multiple_open_streams() -> io::Result<()> {
+    let cursor = Cursor::new(Vec::new());
+    let mut comp = CompoundFile::create(cursor)?;
+
+    // Create a stream and write multiple sectors worth of data into it.
+    let mut stream1 = comp.create_stream("/foo")?;
+    let mut data = Vec::<u8>::new();
+    for i in 1..150 {
+        for j in 0..i {
+            data.push(j);
+        }
+    }
+    assert!(data.len() > comp.version().sector_len() * 2);
+    stream1.write_all(&data)?;
+
+    // Create a second stream and copy the first stream into it.  Having two
+    // open streams at once, and interleaving reads and writes between them,
+    // should work fine.
+    let mut stream2 = comp.create_stream("/bar")?;
+    stream1.rewind()?;
+    let num_bytes = io::copy(&mut stream1, &mut stream2)?;
+    assert_eq!(num_bytes, data.len() as u64);
+
+    // Read the copied data out of the second stream and verify that it matches
+    // the original data.
+    let mut copied = Vec::<u8>::new();
+    stream2.rewind()?;
+    let num_bytes = stream2.read_to_end(&mut copied)?;
+    assert_eq!(num_bytes, data.len());
+    assert_eq!(copied, data);
+    Ok(())
+}
+
+#[test]
+fn drop_compound_file_with_stream_open() -> io::Result<()> {
+    let cursor = Cursor::new(Vec::new());
+    let mut comp = CompoundFile::create(cursor)?;
+    let mut stream = comp.create_stream("/foobar")?;
+    stream.write_all(b"Hello, world!")?;
+    comp.into_inner();
+    let result = stream.flush();
+    assert_eq!(result.unwrap_err().to_string(), "CompoundFile was dropped");
+    Ok(())
+}
+
+//===========================================================================//
